  //		Includes
 //
//
#include <stdio.h>
#include <USBKit.h>
#include <BitmapStream.h>
#include <TranslatorRoster.h>
#include <TranslationUtils.h>
#include <File.h>
#include <NodeInfo.h>


  //		Local includes
 //
//
#include "prototypes.h"
#include "picitem.h"
#include "app.h"
#include "cameracommands.h"


  //		Define global vars
 //
//
unsigned char 				flashdevice[20];
struct usbgadget			*gadgetlist=NULL;


  //		External variables
 //
//
extern class ixd_App		*app;


  //		Check if the new device is an IXUS, init and get the pictures
 //
//
bool ixus_init(USBDevice *dev, bool thumbs)
{
unsigned char			*data;
uint16					length, cnt;
const USBEndpoint 		*iept=NULL;
struct ixus_direntry 	*dirent;

// Initialize the USB device
dev->SetConfiguration(dev->ConfigurationAt(0));

if(!(iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0)))
	{
	fprintf(stderr, "did not get an input endpoint.\n");
	return(FALSE);
	}

if(dev->VendorID() == VENDOR_ID_CANON)
	{
	switch(app->productid = dev->ProductID())
		{
		case PRODUCT_ID_S10:
		case PRODUCT_ID_S20:
		case PRODUCT_ID_S100_US:
		case PRODUCT_ID_S100_EU:
		case PRODUCT_ID_G1:
		
		case PRODUCT_ID_UKNOWN_1:
		case PRODUCT_ID_UKNOWN_2:
		case PRODUCT_ID_UKNOWN_3:
		case PRODUCT_ID_UKNOWN_4:
		case PRODUCT_ID_UKNOWN_5:
		case PRODUCT_ID_UKNOWN_6:
		case PRODUCT_ID_UKNOWN_7:
		case PRODUCT_ID_UKNOWN_8:
		case PRODUCT_ID_UKNOWN_9:

		case PRODUCT_ID_S100V_EU:
		// Begin initialization
		data=app->buf;

		 // IXUS INIT part 1
		//
		data[0]=0;
		length=1;
		if((dev->ControlTransfer(IXUS_READ,				//	request_type
						(length > 1 ? 0x04 : 0x0C),		//	request
						0x55,							//	value
					 	0x00,							//	index
					 	length,							//	length
					 	data)							//	data
						)==-1)
		{
		err_req("Could not initialize the camera (1)");
		fprintf(stderr, "Error in init 1\n");
		return(FALSE);
		}
				
		 // IXUS INIT part 2
		//
		length=0x58;
		for(cnt=0; cnt<length; cnt++) data[cnt]=0;
		if((dev->ControlTransfer(IXUS_READ,				//	request_type
						(length > 1 ? 0x04 : 0x0C),		//	request
						0x1,							//	value
					 	0x00,							//	index
					 	length,							//	length
					 	data)							//	data
						)==-1)
		{
		err_req("Could not initialize the camera (2)");
		fprintf(stderr, "Error in init 2\n");
		return(FALSE);
		}

		 // IXUS INIT part 3
		//
		data=app->buf+0x48;
		length=0x10;
		if((dev->ControlTransfer(IXUS_WRITE,			//	request_type
						(length > 1 ? 0x04 : 0x0C),		//	request
						0x11,							//	value
					 	0x00,							//	index
					 	length,							//	length
					 	data)							//	data
						)==-1)
		{
		err_req("Could not initialize the camera (3)");
		fprintf(stderr, "Error in init 3\n");
		return(FALSE);
		}

		// reset pointer
		data=app->buf;

		 // IXUS INIT part 4
		//
		length=0x44;
		for(cnt=0; cnt<length; cnt++) data[cnt]=0;

		if(iept->BulkTransfer(data, length)==-1)
			{
			err_req("Could not initialize the camera (4)");
			fprintf(stderr, "Error in init 4\n");
			return(FALSE);
			}

		 //	Get the files
		//
		if(thumbs)
			{
			if((dirent=ixus_getdir(dev, data, ixus_flashdevice(dev, data))))
				{
				app->clear_list();
				ixus_getthumbs(dev, data, dirent);
				ixus_freedir(dirent);
				return(TRUE);
				}
			else
				{
				return(FALSE);
				}
			}
		break;

		default:
		return(FALSE);
		break;
		}
	}
else
	{
	return(FALSE);
	}
return(FALSE);
}



 //		Initialize the command block
// 		NOTE: all ints must be sent as small_endian to IXUS
void fillcmdblk(unsigned char *buf, uint32 length, unsigned char cmd1, unsigned char cmd2, uint32 cmd3, uint32 sernum)
{
uint32 l_length=B_HOST_TO_LENDIAN_INT32(length);
uint32 l_cmd3 = B_HOST_TO_LENDIAN_INT32(cmd3);
uint32 l_sernum=B_HOST_TO_LENDIAN_INT32(sernum);

uint16 cnt;

for(cnt=0; cnt < 0x50; cnt++)
	{
	buf[cnt]=0;
	}

buf[3] = ((l_length & 0xff000000)>>24);
buf[2] = ((l_length & 0x00ff0000)>>16);
buf[1] = ((l_length & 0x0000ff00)>>8);
buf[0] = ((l_length & 0x000000ff));

buf[7] = ((l_cmd3 & 0xff000000)>>24);
buf[6] = ((l_cmd3 & 0x00ff0000)>>16);
buf[5] = ((l_cmd3 & 0x0000ff00)>>8);
buf[4] = ((l_cmd3 & 0x000000ff));

buf[0x40] = 0x02;
buf[0x44] = cmd1;
buf[0x47] = cmd2;

buf[0x4b] = ((l_length & 0xff000000)>>24);
buf[0x4a] = ((l_length & 0x00ff0000)>>16);
buf[0x49] = ((l_length & 0x0000ff00)>>8);
buf[0x48] = ((l_length & 0x000000ff));

buf[0x4f] = ((l_sernum & 0xff000000)>>24);
buf[0x4e] = ((l_sernum & 0x00ff0000)>>16);
buf[0x4d] = ((l_sernum & 0x0000ff00)>>8);
buf[0x4c] = ((l_sernum & 0x000000ff));
}


 //		Get a directory listing from the camera
//
struct ixus_direntry *ixus_getdir(USBDevice *dev, unsigned char *buf, unsigned char *path)
{
uint16 				chunk, cnt;
uint32 				length;
struct 				ixus_direntry *result=NULL;
unsigned char 				*dirbuf=NULL;
const USBEndpoint 	*iept;

if(!(iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0)))
	{
	return(NULL);
	}


path=path?path:(unsigned char *)"D:";
fprintf(stderr, "getdir path: %s\n", path);

length=0x10+strlen((char *)path)+1+3;
fillcmdblk(buf, length, 0x0b, 0x11, 0x202, 0x11336677);
buf[0x50]=0x02; // directory recursion
strcpy((char *)buf+0x51, (char *)path);
buf[0x51+strlen((char *)path)+1]=0;
buf[0x51+strlen((char *)path)+2]=0;

length+=0x40;

// Send the command for directory listing
if((dev->ControlTransfer(IXUS_WRITE,			//	request_type
				(length > 1 ? 0x04 : 0x0C),		//	request
				0x10,							//	value
			 	0x00,							//	index
			 	length,							//	length
			 	buf)							//	data
				)==-1)
	{
	fprintf(stderr, "Error in getdir 1\n");
	}

//hexdump(buf, length);

// Check the length of the directory listing
length=0x40;
for(cnt=0; cnt<length; cnt++) buf[cnt]=0;

if(iept)
	{
	if((iept->BulkTransfer(buf, length))==-1)
		{fprintf(stderr, "Error in getdir 2\n");}

	length =(buf[0x09]<<24);
	length|=(buf[0x08]<<16);
	length|=(buf[0x07]<<8);
	length|=(buf[0x06]);
	length=B_LENDIAN_TO_HOST_INT32(length);

	// load the dir listing
	if(length)
		{
		if((dirbuf=(unsigned char *)malloc(length)))
			{
			for(cnt=0; cnt<length;)
				{
				chunk=(((length-cnt)<1024)?(length-cnt):1024);

				if(iept->BulkTransfer(dirbuf+cnt, chunk)==-1)
					{
					fprintf(stderr, "Could not get the directory listing (%d/%ld).\n", cnt, length);
					free(dirbuf);
					return(NULL);
					}
				cnt+=chunk;
				}
			}
		else
			{
			fprintf(stderr, "Could not malloc directory buffer.\n");
			return(NULL);
			}

		result=ixus_parsedir(dirbuf, length);
		free(dirbuf);
		return(result);
		}
	}
else
	{
	fprintf(stderr, "getdir: no input endpoint\n");
	}
return(NULL);
}



 //		Get the name of the flash device in the camera, most likely D:
//
unsigned char *ixus_flashdevice(USBDevice *dev, unsigned char *buf)
{
uint16 cnt;
uint32 length;
const USBEndpoint 	*iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0);	

if(app->productid==PRODUCT_ID_G1)
	{
	fprintf(stderr, "G1 detected, skipping flash device name check.\n");
	return(NULL);
	}


for(cnt=0; cnt<0x60; cnt++) buf[cnt]=0;

fillcmdblk(buf, 0x10, 0x0a, 0x11, 0x202, 0x12345678);
//fillcmdblk(buf, 0x00, 0x0a, 0x11, 0x202, 0x12345678);

length=0x50;

// Send the command for retrieve flash device id
if(dev->ControlTransfer(IXUS_WRITE,				//	request_type
				(length > 1 ? 0x04 : 0x0C),		//	request
				0x10,							//	value
			 	0x00,							//	index
			 	length,							//	length
			 	buf)==-1)						//	data
{
fprintf(stderr, "Could not get flash device name (1).\n");
return(NULL);
}

if(iept)
	{
	length=0x40;
	if(iept->BulkTransfer(buf, length)==-1)
		{
		fprintf(stderr, "Could not get flash device name (2).\n");
		return(NULL);
		}

	length =(buf[0x09]<<24);
	length|=(buf[0x08]<<16);
	length|=(buf[0x07]<<8);
	length|=(buf[0x06]);
	length=B_LENDIAN_TO_HOST_INT32(length);

	if(iept->BulkTransfer(buf, length)==-1)
		{
		fprintf(stderr, "Could not get flash device name (3).\n");
		return(NULL);
		}
	else
		{
		strcpy((char *)flashdevice, (char *)buf);
		return(flashdevice);
		}
	}
else
	{
	fprintf(stderr, "Could not get flash device name (4).\n");
	return(NULL);
	}
}


 //		Download a file from the camera
//
bool ixus_getfile(USBDevice *dev, unsigned char *buf, unsigned char *filename, class pic_Item *item)
{
uint16				cnt;
uint32 				length, progress=0, total;
int					bytesread, systemresult;
char 				localname[80];
char				trancmdline[512];
const USBEndpoint 	*iept=NULL;
BFile				*fh, *tmpfh;
BNodeInfo			*ni;

if((item->pic_Orientation==8) || item->pic_Orientation==7 || (item->pic_Orientation==6))
	{
	strcpy(localname, "p_");
	strcat(localname, (char *)filename+(strlen((char *)filename)-12));
	}
else
	{
	strcpy(localname, (char *)filename+(strlen((char *)filename)-12));
	}


if((fh=new BFile(localname, B_WRITE_ONLY | B_CREATE_FILE )))//| B_FAIL_IF_EXISTS)))
	{
	ssize_t		fherr;
	char		*errmsg;

	if((fherr = fh->InitCheck()) != B_OK)
		{
		switch(fherr)
			{
// Do something better, ask for a new name or some such, or add a number like Tracker
			case B_FILE_EXISTS:
			errmsg="The file already exists.";
			break;

			default:
			errmsg="Could not create the file.";
			}
		err_req(errmsg);
		return(false);
		}
	if((ni=new BNodeInfo(fh)))
		{
		ni->SetType("image/jpeg");
		delete ni;
		}
	SetJPEGFileAttrs(fh, item);
	}

if(!(iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0)))
	{
	err_req("Could not get picture (1)");
	fprintf(stderr, "Error in getfile 0\n");
	delete fh;
	return(false);
	}

length=0x10+0x04+0x04+strlen((char *)filename)+1;

for(cnt=0; cnt<(0x40+length); cnt++) buf[cnt]=0;

fillcmdblk(buf, length, 0x01, 0x11, 0x202, 0x12345678);

buf[0x50+0x00]=0x00;  // download picture
buf[0x50+0x05]=0x14;

strcpy((char *)buf+0x50+0x08, (char *)filename);

length+=0x40;

//hexdump(buf, length);

// Send the command for download
if(dev->ControlTransfer(IXUS_WRITE,				//	request_type
				(length > 1 ? 0x04 : 0x0C),		//	request
				0x10,							//	value
			 	0x00,							//	index
			 	length,							//	length
			 	buf)==-1)						//	data
{
err_req("Could not get picture (2)");
fprintf(stderr, "Error in getfile 1\n");
delete fh;
return(false);
}

length=0x40;
if(iept->BulkTransfer(buf, length)==-1)
	{
	err_req("Could not get picture (3)");
	fprintf(stderr, "Error in getfile 2\n");
	delete fh;
	return(false);
	}

length = ((buf[0x09]<<24) | (buf[0x08]<<16) | (buf[0x07]<<8) | (buf[0x06]));
total=length=B_LENDIAN_TO_HOST_INT32(length);

//fprintf(stderr, "\nTransfer: %10d / %10ld%c", 0, total, 13);

unsigned char 	skip=0;
uint32			collect=0;
if(length)
	{
	while(length)
		{
		progress+=(bytesread=iept->BulkTransfer(buf, (length>DL_SIZE?DL_SIZE:length)));
			
//		fprintf(stderr, "Transfer: %10ld / %10ld%c", progress, total, 13);

		if(bytesread != -1)
			{
			ssize_t		fherr;
			char		*errmsg;

// Errors;
// Should gracefully continue the transfer but just discard of the data
// so the camera doesn't hang up on us.	Don't know of any command to cancel the transfer
			if(( (fherr = fh->Write(buf, bytesread)) != bytesread))
				{
				switch(fherr)
					{
					case B_DEVICE_FULL:
					errmsg="The volume is full.";
					break;

					case B_FILE_EXISTS:
					errmsg="The file already exists.";
					break;
	
					default:
					errmsg="Could not write to the file.";
					}
				err_req(errmsg);
				delete fh;
				return(false);
				}
			length -= bytesread;
			collect+=bytesread;
			if((skip == 32))
				{
				app->updatestatus(collect, NULL);
				collect=0;
				skip=0;
				}
			skip++;
			}
		else
			{
			err_req("Transfer failed");
			fprintf(stderr, "\nError in transfer, please check the camera.\n");
			delete fh;
			return(false);
			}
		}
	app->updatestatus(collect, NULL);
	}

systemresult=-1;
// work in progress 2001-04-19
// use jpegtran to rotate +/- 90 degrees for portrait images
if(item->pic_Orientation==6)
	{
	sprintf(trancmdline, "jpegtran -rotate 90 -outfile %s %s\n", localname+2, localname);
	fprintf(stderr, "\n rotate 90\n");
	systemresult=system(trancmdline);
	}
if(item->pic_Orientation==7)
	{
	sprintf(trancmdline, "jpegtran -rotate 180 -outfile %s %s\n", localname+2, localname);
	fprintf(stderr, "\n rotate 90\n");
	systemresult=system(trancmdline);
	}
if(item->pic_Orientation==8)
	{
	sprintf(trancmdline, "jpegtran -rotate 270 -outfile %s %s\n", localname+2, localname);
	fprintf(stderr, "\n rotate 270\n");
	systemresult=system(trancmdline);
	}
if((item->pic_Orientation==8 || item->pic_Orientation==7 || item->pic_Orientation==6) && (systemresult==0))
	{
	if((tmpfh=new BFile(localname+2, B_READ_ONLY)))
		{
		if(tmpfh->InitCheck() == B_OK)
			{		
			if((ni=new BNodeInfo(tmpfh)))
				{
				ni->SetType("image/jpeg");
				delete ni;
				}
// 2001-08-11 ////////////////  note: tmpfh
			SetJPEGFileAttrs(tmpfh, item);
/////////////////////////////
			}
		}
	delete tmpfh;

	// delete the source file
	BEntry *entry;
	if((entry=new BEntry(localname)))
		{
		entry->Remove();
		delete entry;
		}
	}
//
//

delete fh;
return(true);
}




  //	Set attribute to a file in the camera
 //
//			
//	bit 0	write protected
//		4	directory
//		5	not downloaded
//		7	directory
//
bool ixus_setattr(USBDevice *dev, unsigned char *buf, unsigned char *filename, unsigned char attr)
{
uint16 cnt;
uint32 length, lastslash;
const USBEndpoint 	*iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0);	

if(! iept)
	{
	return(false);
	}

for(cnt=0; cnt<0x60; cnt++) buf[cnt]=0;

length=0x10+0x04+strlen((char *)filename)+1;

fillcmdblk(buf, length, 0x0e, 0x11, 0x201, 0x12345678);
length+=0x40;

buf[0x50]=attr;

for(cnt=0, lastslash=0; filename[cnt]; cnt++)
	{
	buf[0x54+cnt]=filename[cnt];
	if(filename[cnt]=='\\')
		{
		lastslash=cnt;
		}
	}
buf[0x54+cnt]=0;
buf[0x54+lastslash]=0;

// Send the command
if(dev->ControlTransfer(IXUS_WRITE,				//	request_type
				(length > 1 ? 0x04 : 0x0C),		//	request
				0x10,							//	value
			 	0x00,							//	index
			 	length,							//	length
			 	buf)==-1)						//	data
	{
	fprintf(stderr, "Error in setattr.\n");
	return(false);
	}

length=0x54;
if(iept->BulkTransfer(buf, length)==-1)
	{
	fprintf(stderr, "Error in setattr 2.");
	return(false);
	}

return(true);
}




 //		Download a thumbnail from the camera
//
bool ixus_getthumb(USBDevice *dev, unsigned char *buf, unsigned char *filename, class pic_Item *item)
{
uint16				cnt;
uint32 				length, total;
const USBEndpoint 	*iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0);	
BMallocIO			*memfh;
BTranslatorRoster 	*roster = BTranslatorRoster::Default();
BBitmapStream 		stream;

if(!(memfh=new BMallocIO()))
	{
	return(false);
	}

if(! iept)
	{
	delete memfh;
	return(false);
	}

length=0x10+0x04+0x04+strlen((char *)filename)+1;

for(cnt=0; cnt<(0x40+length); cnt++) buf[cnt]=0;

fillcmdblk(buf, length, 0x01, 0x11, 0x202, 0x12345678);

buf[0x50+0x00]=0x01;  // download thumbnail
buf[0x50+0x05]=0x14;
strcpy((char *)buf+0x50+0x08, (char *)filename);
length+=0x40;

// Send the command for download
if(dev->ControlTransfer(IXUS_WRITE,				//	request_type
				(length > 1 ? 0x04 : 0x0C),		//	request
				0x10,							//	value
			 	0x00,							//	index
			 	length,							//	length
			 	buf)==-1)						//	data
{
fprintf(stderr, "Error in getthumb 1\n");
delete memfh;
return(false);
}

length=0x40;
if(iept->BulkTransfer(buf, length)==-1)
	{
	fprintf(stderr, "Error in getthumb 2\n");
	delete memfh;
	return(false);
	}

length =(buf[0x09]<<24);
length|=(buf[0x08]<<16);
length|=(buf[0x07]<<8);
length|=(buf[0x06]);
total=length=B_LENDIAN_TO_HOST_INT32(length);

//fprintf(stderr, "Thumbnail: %ld\n", length);


if(ixus_exifparse(iept, buf, memfh, length, item))
	{
	item->set_Bitmap(BTranslationUtils::GetBitmap(memfh, roster));
	delete memfh;

	// flip the thumbnail if orientation is 6 or 8 
	if(item->pic_Orientation==6)
		{
		item->Rotate(90);
		}
	else if(item->pic_Orientation==8)
		{
		item->Rotate(270);
		}
	}
else
	{
	return(false);
	}

return(true);
}

  //	Delete a file in the camera
 //
//			
uint32 ixus_delete(USBDevice *dev, unsigned char *buf, unsigned char *filename)
{
uint16 cnt;
uint32 length, lastslash, result=0xffffffff;
const USBEndpoint 	*iept;

if(!(iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0)))
	{
	return(result);
	}

for(cnt=0; cnt<0x60; cnt++) buf[cnt]=0;

length=0x10+strlen((char *)filename)+1;

fillcmdblk(buf, length, 0x0d, 0x11, 0x201, 0x12345678);
length+=0x40;

for(cnt=0, lastslash=0; filename[cnt]; cnt++)
	{
	buf[0x50+cnt]=filename[cnt];
	if(filename[cnt]=='\\')
		{
		lastslash=cnt;
		}
	}
buf[0x50+cnt]=0;
buf[0x50+lastslash]=0;

// Send the command
if(dev->ControlTransfer(IXUS_WRITE,				//	request_type
				(length > 1 ? 0x04 : 0x0C),		//	request
				0x10,							//	value
			 	0x00,							//	index
			 	length,							//	length
			 	buf)==-1)						//	data
	{
	fprintf(stderr, "Error in delete.\n");
	return(result);
	}

length=0x54;
if(iept->BulkTransfer(buf, length)==-1)
	{
	fprintf(stderr, "Error in delete 2.");
	return(result);
	}

result =(buf[0x53]<<24);
result|=(buf[0x52]<<16);
result|=(buf[0x51]<<8);
result|=(buf[0x50]);
result=B_LENDIAN_TO_HOST_INT32(result);
//hexdump(buf, length);

return(result);
}


 //		Identify camera
//
bool	ixus_identify(USBDevice *dev, unsigned char *buf)
{
uint16 cnt;
uint32 length;
const USBEndpoint 	*iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0);	


for(cnt=0; cnt<0x60; cnt++) buf[cnt]=0;

fillcmdblk(buf, 0x10, 0x01, 0x12, 0x201, 0x12345678);

length=0x50;

// Send the command to retrieve camera id
if(dev->ControlTransfer(IXUS_WRITE,				//	request_type
				(length > 1 ? 0x04 : 0x0C),		//	request
				0x10,							//	value
			 	0x00,							//	index
			 	length,							//	length
			 	buf)==-1)						//	data
{
fprintf(stderr, "Could not get camera id (1).\n");
return(false);
}


if(iept)
	{
	length=0x9c;
	if(iept->BulkTransfer(buf, length)==-1)
		{
		fprintf(stderr, "Could not get camera id (2).\n");
		return(false);
		}

	strncpy(app->cameraname, (const char *)buf+0x5c, 32);
//	strncpy(app->ownername, (const char *)buf+0x7c, 32);

	int32 srccnt, destcnt=200, state=0;

	for(srccnt=0; buf[srccnt+0x7c] && srccnt<32; srccnt++);

	convert_to_utf8(B_ISO1_CONVERSION,
		(const char *)buf+0x7c,
		&srccnt,
		app->ownername,
		&destcnt,
		&state,
		'?');


	app->ownername[destcnt]=0;
//fprintf(stderr, "ident: (%s) src %ld dest %ld\n", app->ownername, srccnt, destcnt);	
	}
else
	{
	fprintf(stderr, "Could not get camera id (3).\n");
	return(false);
	}
return(true);
}


 //		Setowner
//
bool ixus_setowner(USBDevice *dev, unsigned char *buf, const char *ownername)
{
uint16 cnt;
uint32 length;
char	tempc[40];
const USBEndpoint 	*iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0);	

if(! iept)
	{
	return(false);
	}

for(cnt=0; cnt<0x60; cnt++) buf[cnt]=0;

length=0x10+strlen((char *)ownername)+1;

fillcmdblk(buf, length, 0x05, 0x12, 0x201, 0x12345678);
length+=0x40;

int32 srccnt, destcnt=31, state=0;

for(srccnt=0; ownername[srccnt] && srccnt<200; srccnt++);

convert_from_utf8(B_ISO1_CONVERSION,
	ownername,
	&srccnt,
	tempc,
	&destcnt,
	&state,
	'?');
tempc[destcnt]=0;
//fprintf(stderr, "setowner: (%s) src %ld dest %ld\n", tempc, srccnt, destcnt);	

for(cnt=0; tempc[cnt] && (cnt<31); cnt++)
	{
	buf[0x50+cnt]=tempc[cnt];
	}
buf[0x50+cnt]=0;

// Send the command
if(dev->ControlTransfer(IXUS_WRITE,				//	request_type
				(length > 1 ? 0x04 : 0x0C),		//	request
				0x10,							//	value
			 	0x00,							//	index
			 	length,							//	length
			 	buf)==-1)						//	data
	{
	fprintf(stderr, "Error in setowner.\n");
	return(false);
	}

length=0x54;
if(iept->BulkTransfer(buf, length)==-1)
	{
	fprintf(stderr, "Error in setowner (2).");
	return(false);
	}

return(true);
}


 //		Diskinfo request
//
bool ixus_diskinfo(USBDevice *dev, unsigned char *buf, const char *diskname)
{
uint16 cnt;
uint32 length;
const USBEndpoint 	*iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0);	


if(! iept)
	{
	return(false);
	}


for(cnt=0; cnt<0x60; cnt++) buf[cnt]=0;

length=0x10+strlen((char *)diskname)+1+1;


fillcmdblk(buf, length, 0x09, 0x11, 0x201, 0x12345678);
length+=0x40;


for(cnt=0; diskname[cnt]; cnt++)
	{
	buf[0x50+cnt]=diskname[cnt];
	}
buf[0x50+cnt]='\\';
buf[0x50+cnt+1]=0;



// Send the command
if(dev->ControlTransfer(IXUS_WRITE,				//	request_type
				(length > 1 ? 0x04 : 0x0C),		//	request
				0x10,							//	value
			 	0x00,							//	index
			 	length,							//	length
			 	buf)==-1)						//	data
	{
	fprintf(stderr, "Error in diskinfo.\n");
	return(false);
	}

//////////////////////////////////  STRANGE  /////////////////////////////////////
fprintf(stderr, "------- diskinfo pre bulktran.\n");
//
// There's a 4 second delay here the first time you make the disk info rquest....
//

length=0x5c;
if(iept->BulkTransfer(buf, length)==-1)
	{
	fprintf(stderr, "Error in diskinfo (2).\n");
	return(false);
	}

//
//
fprintf(stderr, "------- diskinfo post bulktran.\n");
//////////////////////////////////////////////////////////////////////////////////


app->totalbytes =(buf[0x57]<<24);
app->totalbytes|=(buf[0x56]<<16);
app->totalbytes|=(buf[0x55]<<8);
app->totalbytes|=(buf[0x54]);
app->totalbytes=B_LENDIAN_TO_HOST_INT32(app->totalbytes);

app->freebytes =(buf[0x5b]<<24);
app->freebytes|=(buf[0x5a]<<16);
app->freebytes|=(buf[0x59]<<8);
app->freebytes|=(buf[0x58]);
app->freebytes=B_LENDIAN_TO_HOST_INT32(app->freebytes);

return(true);
}



  //		powerinfo request
 //
//
bool ixus_powerinfo(USBDevice *dev, unsigned char *buf)
{
uint16 cnt;
uint32 length;
const USBEndpoint 	*iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0);	

if(! iept)
	{
	return(false);
	}

for(cnt=0; cnt<0x60; cnt++) buf[cnt]=0;

length=0x10;
fillcmdblk(buf, length, 0x0a, 0x12, 0x201, 0x12345678);
length+=0x40;

// Send the command
if(dev->ControlTransfer(IXUS_WRITE,				//	request_type
				(length > 1 ? 0x04 : 0x0C),		//	request
				0x10,							//	value
			 	0x00,							//	index
			 	length,							//	length
			 	buf)==-1)						//	data
	{
	fprintf(stderr, "Error in powerinfo.\n");
	return(false);
	}

//fprintf(stderr, "diskinfo.\n");
//hexdump(buf, length);

length=0x58;
if(iept->BulkTransfer(buf, length)==-1)
	{
	fprintf(stderr, "Error in powerinfo (2).\n");
	return(false);
	}

//fprintf(stderr, "powerinfo.\n");
//fprintf(stderr, "powerinfo: %02x %02x\n", buf[0x54], buf[0x57]);
//hexdump(buf, length);

app->powerstatus=0;
app->powerstatus|=((buf[0x54]==0x04)?1:0);
app->powerstatus|=((buf[0x57]==0x10)?1:0);

return(true);
}


#define UL_SIZE 0x300	// do not work if its higher than 0x300


  //		Upload a file to the camera
 //
//
bool ixus_sendfile(USBDevice *dev, unsigned char *buf, const char *filename)
{
uint16				cnt;
uint32 				length;
const USBEndpoint 	*iept=NULL, *oept=NULL;
BFile				*fh;
off_t				filesize=0;
uint16				len1;
uint32 				l_sernum=0x12345678;
uint32				offset=0;
char				localname[100];

strcpy(localname, "D:\\DCIM\\112CANON\\IMG_1205.JPG");


if((fh=new BFile(filename, B_READ_ONLY)))
	{
	if(fh->InitCheck() != B_OK)
		{
		err_req("Could not open the file");
		return(false);
		}
	fh->GetSize(&filesize);
	}

for(cnt=strlen((const char *)filename); cnt && filename[cnt]!='/'; cnt--);
if(filename[cnt]=='/') cnt++;

sprintf((char *)buf, "Uploading %s.", (const char *)filename+cnt);
app->openstatus(filesize, (char *)buf);
app->updatestatus(0, (char *)buf);

if(!(iept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0)))
	{
	app->closestatus();
	err_req("Could not send picture (1)");
	delete fh;
	return(false);
	}

if(!(oept=dev->ActiveConfiguration()->InterfaceAt(0)->EndpointAt(1)))
	{
	app->closestatus();
	err_req("Could not send picture (2)");
	delete fh;
	return(false);
	}

fprintf(stderr, "oept: %s %s\n", oept->IsBulk()?"bulk":"not bulk", oept->IsOutput()?"output":"no output");

if(oept->IsInput() || (!oept->IsBulk()))
	{
	app->closestatus();
	err_req("Error: no output bulk endpoint!!");
	delete fh;
	return(false);
	}



// Transfer the file
ssize_t		datalength;
char		readbuffer[UL_SIZE];

while(1)
	{
	// read in part of the file in the buffer
	datalength=fh->Read(readbuffer, UL_SIZE);

	if(datalength < 0)
		{
		app->closestatus();
		delete fh;
		return(false);
		}
	else if(datalength == 0)
		{
		app->closestatus();
		delete fh;
		return(true);
		}
	len1=0x1c+strlen(localname)+1+datalength;




// first control transfer
	for(cnt=0; cnt<0x40; cnt++) buf[cnt]=0;
	buf[4]=0x03;
	buf[5]=0x02;
	buf[7] = (((len1+0x40) & 0xff00)>>8);
	buf[6] = (((len1+0x40) & 0x00ff));
	length=0x40;
	if(dev->ControlTransfer(IXUS_WRITE,				//	request_type
					(length > 1 ? 0x04 : 0x0C),		//	request
					0x10,							//	value
				 	0x00,							//	index
				 	length,							//	length
			 		buf)==-1)						//	data
		{
		app->closestatus();
		err_req("Could not send picture (3)");
		delete fh;
		return(false);
		}

// reply from the transfer
	length=0x40;
	if(iept->BulkTransfer(buf, length)==-1)
		{
		app->closestatus();
		err_req("Could not send picture (4)");
		delete fh;
		return(false);
		}

// send data
	for(cnt=0; cnt<0x80; cnt++) buf[cnt]=0;
	strcpy((char *)buf+0x5c, localname);

	buf[0x01] = ((len1 & 0xff00)>>8);
	buf[0x00] = ((len1 & 0x00ff));

	buf[0x04] = 0x03;
	buf[0x05] = 0x04;

	buf[0x40] = 0x02;
	buf[0x44] = 0x03;
	buf[0x47] = 0x11;

	buf[0x49] = ((len1 & 0xff00)>>8);
	buf[0x48] = ((len1 & 0x00ff));

	buf[0x4f] = ((l_sernum & 0xff000000)>>24);
	buf[0x4e] = ((l_sernum & 0x00ff0000)>>16);
	buf[0x4d] = ((l_sernum & 0x0000ff00)>>8);
	buf[0x4c] = ((l_sernum & 0x000000ff));

	buf[0x50] = 0x02;

	buf[0x57] = ((offset & 0xff000000)>>24);
	buf[0x56] = ((offset & 0x00ff0000)>>16);
	buf[0x55] = ((offset & 0x0000ff00)>>8);
	buf[0x54] = ((offset & 0x000000ff));

	buf[0x59] = ((datalength & 0xff00)>>8);
	buf[0x58] = ((datalength & 0x00ff));

	memcpy(buf+(0x5c+strlen(localname)+1), readbuffer, datalength);

	if(oept->BulkTransfer(buf, (len1+0x40))==-1)
		{
		app->closestatus();
		err_req("Could not send picture (5)");
		delete fh;
		return(false);
		}

	offset+=datalength;

	app->updatestatus(datalength, NULL);

// reply for data
	length=0x5c;
	if(iept->BulkTransfer(buf, length)==-1)
		{
		app->closestatus();
		err_req("Could not send picture (6)");
		delete fh;
		return(false);
		}

	}

app->closestatus();
delete fh;
return(true);
}


 //		Download all the thumbnails from the camera
//
void ixus_getthumbs(USBDevice *dev, unsigned char *buf, struct ixus_direntry *dirent)
{
unsigned char 					pathbuf[255];
struct	ixus_direntry 	*thisentry=dirent;
uint32 					lastslash, cnt;
class	pic_Item		*item;

pathbuf[0]=0;

while(thisentry)
	{
	if(thisentry->name[1] == ':')
		{
		strcpy((char *)pathbuf, (char *)thisentry->name);
		}
	if(!strncmp((char *)thisentry->name, ".\\", 2))
		{
		strcpy((char *)pathbuf+strlen((char *)pathbuf), (char *)thisentry->name+1);
		}

	if(!strncmp((char *)thisentry->name, "..", 2))
		{
		for(lastslash=0, cnt=0; pathbuf[cnt]; cnt++)
			{
			if(pathbuf[cnt]=='\\')
				lastslash=cnt;
			}
		pathbuf[lastslash]=0;
		}
	if((cnt=strlen((char *)thisentry->name))>4)
	if(!(strcmp((const char *)thisentry->name+(cnt-4), ".JPG")))
		{
		strcpy((char *)pathbuf+strlen((char *)pathbuf), (char *)"\\");
		strcpy((char *)pathbuf+strlen((char *)pathbuf), (char *)thisentry->name);
		
// work in progress 2001-04-16		
//
		if((item=new pic_Item()))
			{
			item->set_XRes(0);
			item->set_YRes(0);
			item->set_Orientation(0);

			ixus_getthumb(dev, buf, pathbuf, item);

			item->set_Name((char *)pathbuf);
			item->set_Attrs(thisentry->attr);
			item->set_Size(thisentry->size);
			item->set_Date(thisentry->date);
			app->add_item(item);
			}
//
//
		for(lastslash=0, cnt=0; pathbuf[cnt]; cnt++)
			{
			if(pathbuf[cnt]=='\\')
				lastslash=cnt;
			}
		pathbuf[lastslash]=0;
		}
	thisentry=thisentry->next;
	}

	app->sort_list();
}




 //		Convert the directory listing from the camera to a linked list
//
struct ixus_direntry *ixus_parsedir(unsigned char *buf, uint32 length)
{
unsigned char 		*data=buf;
unsigned char		attr=0, *name=(unsigned char *)"";
uint32		size=1, date=0;
struct ixus_direntry *lastdir=0;
struct ixus_direntry *lastentry=0, *thisentry;
struct ixus_direntry *firstentry=0;

fprintf(stderr, "DIRECTORY LISTING:\n");

while(attr || size || date || name[0])
	{
	attr=data[0];

	size =(data[0x05]<<24);
	size|=(data[0x04]<<16);
	size|=(data[0x03]<<8);
	size|=(data[0x02]);
	size=B_LENDIAN_TO_HOST_INT32(size); // not needed

	date =(data[0x09]<<24);
	date|=(data[0x08]<<16);
	date|=(data[0x07]<<8);
	date|=(data[0x06]);
	date=B_LENDIAN_TO_HOST_INT32(date); // ditto

	name=data+0xa;

	if(attr || size || date || name[0])
		{
		if((thisentry=(struct ixus_direntry *)malloc(sizeof(struct ixus_direntry)+strlen((char *)name))))
			{
			if(lastentry)
				{
				lastentry->next=thisentry;
				}
			thisentry->next=NULL;
			thisentry->parent=lastdir;
			thisentry->size=size;
			thisentry->date=date;
			thisentry->attr=attr;
			strcpy((char *)thisentry->name, (char *)name);

			if(!firstentry)
				{
				firstentry=thisentry;
				}

			if(((!strncmp((char *)name, ".\\", 2)) || (name[1]==':')) && size==0)
				{
				lastdir=thisentry;
				}

			if((!strncmp((char *)name, "..", 2)) && size==0)
				{
				lastdir=lastdir?lastdir->parent:NULL;
				}
			lastentry=thisentry;
			}
		else
			{
			fprintf(stderr, "parsedir: could not malloc direntry.\n");			
			}
		}
 
 	fprintf(stderr, "%-13s %12ld [%c%c%c%c] %26s", name, size, 
					attr&1?'P':'-',
					attr&16?'S':'-',
					attr&32?'-':'D',
					attr&128?'R':'-',
					date?asctime(gmtime((const time_t *)&date)):"\n");

	data+=10+strlen((char *)name)+1;
	}
return(firstentry);
}



 //		Free the memory allocated for the directory list
//
void ixus_freedir(struct ixus_direntry *dirent)
{
struct ixus_direntry *thisentry=dirent, *nextentry;

while(thisentry)
	{
	nextentry=thisentry->next;
	free(thisentry);
	thisentry=nextentry;
	}

}


