
  //			Includes
 //
//
#include <stdio.h>
#include <USBKit.h>
#include <File.h>


  //		Local includes
 //
//
#include "picitem.h"
#include "exifparse.h"
#include "cameracommands.h"


  //		defines
 //
//
#define uint16 unsigned short
#define uint32 unsigned long


  //	Uncomment to get debug output
 //
//
//#define exif_debug


  //		Function prototypes
 //
//
int					ixus_exifparse(const class USBEndpoint *iept, unsigned char *buf, class BMallocIO *memfh, uint32 length, class pic_Item *item);
const char * 		exiftagname(struct exiftag taglist[], unsigned int tag);
void 				ie_insertsort(struct ifd_entry *entry);
struct ifd_entry * 	ie_create(unsigned short tagnum, unsigned short format, unsigned long size, unsigned long offset);
void 				ie_ready_push(void);
void 				ie_print(struct ifd_entry *entry);
void 				SetJPEGFileAttrs(class BFile *fh, class pic_Item *item);


  //		Global data
 //
//
unsigned char sizeary[13]={0, 1, 1, 2, 4, 8, 1, 1, 2, 4, 8, 4, 8};
struct ifd_entry	*first_entry=NULL, *ready_entry=NULL;


  //		ie_exifparse()
 //
//
int ixus_exifparse(const USBEndpoint *iept, unsigned char *buf, BMallocIO *memfh, uint32 length, class pic_Item *item)
{
int			bytesread, cnt;

if(length)
	{
	uint32			offset=0, skip=0, fill=0, fillcnt=0;
	uint32			ifd_numcomp, ifd_data, marker=0, markercnt=0;
	uint32  		nextifd=0, thumbifd=0, thumbdata=0, thumbsize=0, makernote=0, interoper=0;
	uint16			ifd_numdir, ifd_tag, ifd_format;
	char			state=0, byteorder='I';
	const char 		*tmps;
	struct exiftag	*currenttags=exif_tags;

	while(length)
		{
		bytesread=iept->BulkTransfer(buf, (length>DL_SIZE?DL_SIZE:length));

		if(bytesread != -1)
			{
			for(cnt=0; cnt<bytesread; cnt++, offset++)
				{
				marker=marker<<8;
				marker|=buf[cnt];
				markercnt++;

				if(skip)
					{
					if(!(--skip))
						{
						markercnt=0;
						}
					if(first_entry)
						{
//						while(first_entry->offset < offset)
//							{
//							ie_ready_push();
//							}
						if(first_entry->offset==offset)
							{
							fill=(first_entry->size)*sizeary[first_entry->format<13?first_entry->format:ifdf_undef];
							fillcnt=0;
//							fprintf(stderr, "\n%08x %08x fill==%d\n", offset, first_entry->format, fill);
//							ie_ready_push();
							}
						}
					if(fill)
						{
//						fprintf(stderr, "%c", buf[cnt]);
						ready_entry->data[fillcnt++]=buf[cnt];
						fill--;
						if(!fill)
							{
//							fprintf(stderr, "\"\n");
							}
						}
					}
				else
					{
					switch(state)
						{
						case ExifState_init:
						if(markercnt==2)
							{
							if(marker==JPG_SOI)
								{
								marker=markercnt=0;
								state=ExifState_exif_init;
#ifdef exif_debug
								fprintf(stderr, "SOI\n");
#endif

								}
							else
								return(0);
							}
						break;

						case ExifState_exif_init:
						if(markercnt==2)
							{
							if(marker==JPG_APP1)
								{
								marker=markercnt=0;
								state=ExifState_exif_header;
								skip=2;
#ifdef exif_debug
								fprintf(stderr, "APP1\n");
#endif
								}
							else
								{
#ifdef exif_debug
								fprintf(stderr, "%04lx\n", marker);
#endif
								}
							}
						if(markercnt==4)
							{
							skip=B_BENDIAN_TO_HOST_INT32(marker)-2; // BENDIAN?
#ifdef exif_debug
							fprintf(stderr, "Skipping %08lx bytes\n", skip);
#endif
							marker=markercnt=0;
							state=0;
							}
						break;

						case ExifState_exif_header:
						if(buf[cnt] != Exif_header[markercnt-1])
							{
#ifdef exif_debug
							fprintf(stderr, "exif no match '%c' != '%c' (%ld)\n", buf[cnt], Exif_header[markercnt-1], markercnt-1);
#endif
							return(0);
							}
						if(markercnt==6)
							{
#ifdef exif_debug
							fprintf(stderr, "Exif header OK\n");
#endif
							state=ExifState_tiff_header;
							marker=markercnt=0;
							}
						break;
						
						case ExifState_tiff_header:
						offset=0;
						byteorder=buf[cnt];
						skip=7;
						marker=markercnt=0;
						ifd_numdir=0;
						state=ExifState_ifdnum;
						break;

						case ExifState_ifdnum:
						if(markercnt==2)
							{
							ifd_numdir=swapm16(marker);
#ifdef exif_debug
							fprintf(stderr, "numdir %08x marker %08lx\n", ifd_numdir, marker);
#endif
							marker=markercnt=0;
							state=ExifState_ifdentry;
							}
						break;

						case ExifState_ifdentry:
						switch(markercnt)
							{
							case 2:
							ifd_tag=swapm16(marker);
							if(!(tmps=exiftagname(currenttags, ifd_tag)))
								{
//								fprintf(stderr, "error: unknown tag\n");
//								return(0);
								}
#ifdef exif_debug
							fprintf(stderr, "tag %04x: %s", ifd_tag, tmps?tmps:"unknown tag");
#endif
							marker=0;
							break;

							case 4:
							ifd_format=swapm16(marker);
							tmps=exiftagname(&exif_formats[0], ifd_format);
#ifdef exif_debug
							fprintf(stderr, "  (%s)", tmps?tmps:"unknown format");
#endif
							marker=0;
							break;

							case 8:
							ifd_numcomp=swapm(marker);
#ifdef exif_debug
							fprintf(stderr, " size: %ld", ifd_numcomp);
#endif
							marker=0;
							break;

							case 12:
							ifd_data=swapm(marker);
#ifdef exif_debug
							fprintf(stderr, " value: %08lx", ifd_data);
#endif
							if(ifd_format==4 && ifd_numcomp==1)
								{
								if(ifd_tag == 0x8769) // ExifOffset
									{
									nextifd=ifd_data;
									}
								if(ifd_tag == 0x201) // JpegIFOffset
									{
									thumbdata=ifd_data;
									}
								if(ifd_tag == 0x202) // JpegIFLength
									{
									thumbsize=ifd_data;
									}
								if(ifd_tag == 0xA005) // InteroperabilityOffset
									{
									interoper=ifd_data;
									}
								}
							if(ifd_format==3 && ifd_numcomp==1)
								{
								if(ifd_tag == 0xA002) // ExifImageWidth
									{
									item->set_XRes(ifd_data);
									}
								if(ifd_tag == 0xA003) // ExifImageHeight
									{
									item->set_YRes(ifd_data);
									}
								if(ifd_tag == 0x9209) // Flash
									{
									item->set_Flash(ifd_data);
									}
								if(ifd_tag == 0x112) // Orientation
									{
									item->set_Orientation(ifd_data);
									}
								}
							if(ifd_format==7)
								{
								if(ifd_tag == 0x927C) // MakerNote
									{
									makernote=ifd_data;
									}
								}
							if((sizeary[ifd_format<13?ifd_format:ifdf_undef]*ifd_numcomp)>4)
								if(ifd_format!=ifdf_undef){
/*								if((entry=ie_create(ifd_tag, ifd_format, ifd_numcomp, ifd_data)))
									{
									ie_insertsort(entry);
									fprintf(stderr, "\ninsertsort format: %08x %08x\n", ifd_format, ifd_numcomp);
									}
*/								}
#ifdef exif_debug
							fprintf(stderr, "\n");
#endif
							marker=markercnt=0;
							ifd_numdir--;
							if(! ifd_numdir)
								{
								state=ExifState_ifdoffset;
								}
							}
						break;

						case ExifState_ifdoffset:
						if(markercnt==4)
							{
							if(makernote)
								{
								skip=(makernote-offset)-1;
								makernote=0;
								currenttags=mn_canon_tags;
								state=ExifState_ifdnum;
								}
							else
							if(interoper)
								{
								skip=(interoper-offset)-1;
								interoper=0;
								currenttags=exif_itags;
								state=ExifState_ifdnum;
								}
							else
							if(nextifd)
								{
								skip=(nextifd-offset)-1;
								nextifd=0;
								thumbifd=swapm(marker);
								currenttags=exif_tags;
								state=ExifState_ifdnum;
								}
							else if(thumbifd)
								{
								skip=(thumbifd-offset)-1;
								thumbifd=0;
								currenttags=exif_tags;
								state=ExifState_ifdnum;
								}
							else if(thumbdata)
								{
#ifdef exif_debug
								fprintf(stderr, "thumbdata %08lx\n", thumbdata);
#endif
								skip=(thumbdata-offset)-1;
								thumbdata=0;
								state=ExifState_thumbdata;
								}
							else
								{
#ifdef exif_debug
								fprintf(stderr, "no thumbdata\n");
#endif
								state=100;
								}
#ifdef exif_debug
							fprintf(stderr, "offset %08lx skip %08lx marker %08lx\n", offset, skip, marker);
#endif
							marker=markercnt=0;
							}
						break;

						case ExifState_thumbdata:
						if(!(thumbsize))
							{
#ifdef exif_debug
							fprintf(stderr, "got thumbnail data\n");
#endif
							return(1);
							}
						else
							{
							memfh->Write((const void *)(buf+cnt), bytesread-cnt);
							thumbsize-=(bytesread-cnt);
							cnt=bytesread;
							}
						break;

						default:
#ifdef exif_debug
						fprintf(stderr, "Unknown state\n");
#endif
						return(0);
						}
					}
				}

			length -= bytesread;
			}
		else
			{
			return(0);
			}
		}
	}
return(1);
}


  //		ie_print()
 //
//
void ie_print(struct ifd_entry *entry)
{
unsigned long	tmpsh1=0, tmpsh2=0;
unsigned long	cnt, cnt2;

if(!entry)
	{
	return;
	}
switch(entry->format)
	{
	case ifdf_ubyte:
	break;

	case ifdf_string:
	fprintf(stderr, "string: \"%s\"\n", entry->data);
	break;

	case ifdf_ushort:
	fprintf(stderr, "ushort\n");
	for(cnt2=0; cnt2<entry->size; cnt2++)
		{
		tmpsh1=0;
		for(cnt=2; cnt; cnt--)
			{
			tmpsh1=tmpsh1<<8;
			tmpsh1|=entry->data[(cnt2*2)+(cnt-1)];
			}
		fprintf(stderr, "ushort: %04lx\n", tmpsh1);
		}
	break;

	case ifdf_ulong:
	for(cnt2=0; cnt2<entry->size; cnt2++)
		{
		tmpsh1=0;
		for(cnt=4; cnt; cnt--)
			{
			tmpsh1=tmpsh1<<8;
			tmpsh1|=entry->data[(cnt2*4)+(cnt-1)];
			}
		fprintf(stderr, "ulong: %08lx\n", tmpsh1);
		}
	break;

	case ifdf_urat:
	for(cnt=4; cnt; cnt--)
		{
		tmpsh1=tmpsh1<<8;
		tmpsh1|=entry->data[cnt-1];
		}
	for(cnt=7; cnt>3; cnt--)
		{
		tmpsh2=tmpsh2<<8;
		tmpsh2|=entry->data[cnt];
		}
	fprintf(stderr, "urat: %ld/%ld\n", tmpsh1, tmpsh2);
	break;


	case ifdf_srat:
	for(cnt=4; cnt; cnt--)
		{
		tmpsh1=tmpsh1<<8;
		tmpsh1|=entry->data[cnt-1];
		}
	for(cnt=7; cnt>3; cnt--)
		{
		tmpsh2=tmpsh2<<8;
		tmpsh2|=entry->data[cnt];
		}
	fprintf(stderr, "srat: %ld/%ld\n", tmpsh1, tmpsh2);
	break;

/*	case ifdf_sbyte:
	break;

	case ifdf_undef:
	break;

	case ifdf_sshort:
	break;

	case ifdf_slong:
	break;

	case ifdf_float:
	break;

	case ifdf_dfloat:
	break;
*/
	default:
	fprintf(stderr, "unknown: %08x/\n", entry->format);
	}
}


  //		ie_insertsort()
 //
//
void ie_insertsort(struct ifd_entry *entry)
{
struct ifd_entry *next;

if((next=first_entry))
	{
	while(next)
		{
		if(entry->offset <= next->offset)
			{
			if(!(entry->prev=next->prev))
				{
				first_entry=entry;
				}
			entry->next=next;

			next->prev=entry;
			return;
			}
		else if(!(next->next))
			{
			next->next=entry;
			entry->prev=next;
			entry->next=NULL;
			return;
			}
		next=next->next;
		}
	}
else
	{
	first_entry=entry;
	entry->prev=NULL;
	entry->next=NULL;
	}
}


  //		ie_create()
 //
//
struct ifd_entry *ie_create(unsigned short tagnum, unsigned short format, unsigned long size, unsigned long offset)
{
struct ifd_entry *entry;
format=format<13?format:ifdf_undef;

if((entry=(struct ifd_entry *)malloc(sizeof(struct ifd_entry)+(size*sizeary[format]))))
	{
	entry->prev=NULL;
	entry->next=NULL;
	entry->tagnum=tagnum;
	entry->format=format;
	entry->size=size;
	entry->offset=offset;
	}
return(entry);
}


  //		ie_ready_push()
 //
//
void ie_ready_push(void)
{
struct ifd_entry *entry;

if((entry=first_entry))
	{
	if((first_entry=entry->next))
		{
		first_entry->prev=NULL;
		}
	else
		{
		first_entry=NULL;
		}
	entry->prev=NULL;
	entry->next=ready_entry;
	if(ready_entry)
		{
		ready_entry->prev=entry;
		}
	ready_entry=entry;
	}
}


  //		exiftagname()
 //
//
const char * exiftagname(struct exiftag taglist[], unsigned int tag)
{
uint32	cnt;

for(cnt=0; taglist[cnt].name; cnt++)
	{
	if(taglist[cnt].value==tag)
		{
		return(taglist[cnt].name);
		}
	}

return(NULL);
}


  //		SetJPEGFileAttrs()
 //
//
void SetJPEGFileAttrs(class BFile *fh, class pic_Item *item)
{
int32		width=(item->pic_Orientation==1 || item->pic_Orientation==7) ? item->pic_Xres : item->pic_Yres;
int32		height=(item->pic_Orientation==1 || item->pic_Orientation==7) ? item->pic_Yres : item->pic_Xres;
bool		flash=item->pic_Flash?true:false;
bigtime_t	time=mktime(gmtime((const time_t *)(&(item->pic_Date))));

fh->WriteAttr("GRAFX:Width", B_INT32_TYPE, 0, (const void *)&width, sizeof(int32));
fh->WriteAttr("GRAFX:Height", B_INT32_TYPE, 0, (const void *)&height, sizeof(int32));
fh->WriteAttr("EXIF:Flash", B_BOOL_TYPE, 0, (const void *)&flash, sizeof(bool));
fh->WriteAttr("EXIF:Date", B_TIME_TYPE, 0, (const void *)&time, sizeof(time));
}


